import java.io.*;  // For BufferedWriter
import java.util.*; // For LinkedList implementation

/**
 * Class that implements a general Simulation
 **/
public class Simulation {
  
  ////////// Fields /////////////////
  /** Linked lists for tracking all agents in the simulationh */
  protected LinkedList agents = new LinkedList();

  /** World that the simulation runs */
  protected World world;
  
  /** A BufferedWriter for writing to */
  public BufferedWriter output;
  
  /** FrameSequence, if we're doing an animation */
  public FrameSequence frames;
  
  /** A public Frame for setUp (and others) to access, if desired. */
  public Picture frame;
  
  ///// Constructors //////////////
  /**
   * Constructor to set output to null
   **/
  public Simulation() {
    // By default, don't write to a file.
    output = null;
    // And there is no FrameSequence
    frames = null;
  }
  
  //////// Methods ///////////////
    
  /** Accessor for all agents */
  public LinkedList getAgents(){return agents;}
  
  /** 
   * Add this actor to the agents list
   * @param newOne agent to add to the list
   **/
  public void add(Agent newOne){
    agents.add(newOne);
  }
  
  /** 
   * Remove this actor from the agents list
   * @param newOne agent to remove to the list
   **/
  public void remove(Agent newOne){
    agents.remove(newOne);
  }

  /**
   * setUp the simulation.
   * Subclasses should call this (via super)
   * and create agents.
   **/
  public void setUp(){
    if (frames == null){
      // Set up the World
      world = new World();
      world.setAutoRepaint(false);}
  }
  
  /** 
   * openFrames -- create the FrameSequence for storing 
   * frames of the simulation
   **/
  public void openFrames(String directory){
    frames = new FrameSequence(directory);
  }
  
  /**
   * Open the input file and set the BufferedWriter to speak to it.
   **/
  public void openFile(String filename){
   // Try to open the file
    try {
      
      // create a writer 
      output = new BufferedWriter(new FileWriter(filename));
      
    } catch (Exception ex) {
      System.out.println("Trouble opening the file " + filename);
      // If any problem, make it null again
      output = null;
    }

  }
  
  /**
   * frameSetup() -- Something you want to do each frame?
   **/
  public void frameSetup() {
  }
  
  /**
   * Ask all agents to run for the number of input 
   * steps
   **/
  public void run(int timeRange)
  {
    // For storing the current agent
    Agent current = null;
    
    
    // Set up the simulation
    this.setUp();
    
    // loop for a set number of timesteps 
    for (int t = 0; t < timeRange; t++)
    {
          if (frames != null){
            // For storing the current frame, if there is a FrameSequence
            frame = new Picture(640,480);
            this.frameSetup();
          }

      // loop through all the agents, and have them
      // act()
      for (int index=0; index < agents.size(); index++) {
        current = (Agent) agents.get(index);
        current.act();
      }
     // Could separate acting from updating...
      
      if (frames==null){
        // repaint the world to show the movement
        world.repaint();
      } else {
        // Add the frame to the sequence
        frames.addFrame(frame);
      }
        
      
      // Do the end of step processing
      this.endStep(t);
    
      // Wait for one second
      //Thread.sleep(1000);
    }
    
  }
  
  /**
   * Run for a default of 50 steps
   **/
  public void run(){
    this.run(50);
    this.closeFile();
  }
  
  /**
   * End of step processing
   * @param t timestep count
   **/
  public void endStep(int t){
    // Let's figure out where we stand...
    System.out.println(">>> Timestep: "+t);
      
    // If we have an open file, write the counts to it
    if (output != null) {
      // Try it
      try{
        output.write(lineForFile());
        output.newLine();
      } catch (Exception ex) {
        System.out.println("Couldn't write the data!");
        System.out.println(ex.getMessage());
        // Make output null so that we don't keep trying
        output = null;
        }
      }
  }  // endStep()
  
  /**
   * lineForFile -- string to write to file on each
   * timeStep.  Will probably be overridden in subclsses.
   **/
  public String lineForFile(){
    // Get the size (an int), make it an Integer, 
    // in order to turn it into a string. (Whew!)
    return (new Integer(agents.size())).toString();
  }
  
  /**
   * close the file and clear the output stream
   **/
  public void closeFile(){
    // If we have an open file, close it and null the variable
    if (output != null){
      try{
        output.close();}
      catch (Exception ex)
      {System.out.println("Something went wrong closing the file");}
      finally {
        // No matter what, mark the file as not-there
        output = null;}
    }
  } // closeFile()
  
}
